(in-package :cl-user)

(defpackage :if-basic-lib
  (:use :common-lisp)
  (:export :with-gen-syms :once-only :defsyn
	   :*space-chars* :*dictionary* :*dict-index* :*instream*
	   :*outstream* :*verbs* :*tokens* :*allobjects* :*tokenpar*
	   :*action* :*args* :*noun* :*second* :before :after :self
	   :*after*
	   :addword :word2dic :addword2dic
	   :split-to-words :sprint
	   :parser :description :article 
	   :glance :in-scope
	   :initnames :addnames
	   :read-property :read-property-string :read-property-number
	   :read-property-integer :read-property-object :read-property-execute
	   :read-property-other :read-property-list :exec 
	   :abstractobject :name :names :parent :children :flags
	   :initflags :add-flags :has :hasnt :-> :give
	   :ifclass :object :defaction
	   :move :rmv
	   :verb :extend-verb :extend-verb-first
	   :extend-verb-only :extend-verb-only-first
	   :deftoken :string== :matchp :!last!
	   :in :objectloop :provides
	   :wordlist :tokenlist
	   :nosuchword :nosuchword-word
	   :parse-command :unknown-verb :run-action :run-action-after
	   :turn-passing :pretty-string :*textwidth* :*offset* 
	   :ignore-newlines :newline :freshline :put-word :outprinc
	   :destroy))

(in-package :if-basic-lib)

(defmacro with-gen-syms ((&rest names) &body body)
  `(let ,(loop for n in names collect `(,n (make-symbol ,(string n))))
     ,@body))

(defmacro once-only ((&rest names) &body body)
  (let ((gensyms (loop for n in names collect (gensym (string n)))))
    `(let (,@(loop for g in gensyms collect `(,g (gensym))))
      `(let (,,@(loop for g in gensyms for n in names collect ``(,,g ,,n)))
        ,(let (,@(loop for n in names for g in gensyms collect `(,n ,g)))
           ,@body)))))

(defun ignore-warning (condition)
   (declare (ignore condition))
   (muffle-warning))

(defun as-keyword (sym) (intern (string sym) :keyword))

(defparameter *space-chars* #(#\Space #\Newline #\Tab))
(defparameter *dictionary* (make-hash-table :test #'equal))
(defparameter *dict-index* 0)
(defparameter *outstream* *standard-output*)
(defparameter *instream* *standard-input*)
(defparameter *verbs* (make-hash-table :test #'eql))
(defparameter *tokens* (make-hash-table :test #'eql))
(defparameter *allobjects* nil)
(defparameter *tokenpar* :unspecified)
(defparameter *action* nil)
(defparameter *args* nil)
(defparameter *noun* nil)
(defparameter *second* nil)
(defparameter *after* nil) ;;Whether the run-action-after was called during
                           ;;the last action and wasn't interrupted  - this
                           ;;is necessary for indirect action processing
                           ;;(like put-on/in and reverse)
;;Text printer parameters
(defparameter *textwidth* 72)
(defparameter *offset* 0)

(defstruct patternlist value)

(defmacro defsyn (name func)
  `(defmacro ,name (&rest args) (apply ,func args)))

(defun newline (&optional (stream *outstream*))
  (setf *offset* 0) (terpri stream))

(defun freshline (&optional (stream *outstream*))
  (setf *offset* 0) (fresh-line stream))

(defun outprinc (str &optional (stream *outstream*))
  (princ str stream) (incf *offset* (length str)))

(defun ignore-newlines (str)
  (let (spaceflag (countspaces 0))
    (with-output-to-string (out)
      (loop for c across str
	    when (char= c #\Space) do (incf countspaces)
	    else when (char= c #\Newline) do (setf spaceflag t countspaces 0)
	                                     (princ #\Space out)
	    else do (unless spaceflag (loop for i from 1 to countspaces
					    do (princ #\Space out)))
	            (princ c out) (setf spaceflag nil countspaces 0))
      (loop for i from 1 to countspaces do (princ #\Space out)))))
    
(defun put-word (word stream)
  (let ((wordlen (length word)))
    (if (<= (+ *offset* wordlen 1) *textwidth*)
	(progn (princ word stream) (princ #\Space stream)
	       (incf *offset* (1+ wordlen)))
	(progn (newline stream) (princ word stream) 
	       (princ #\Space stream)
	       (incf *offset* (1+ wordlen))))))

(defun pretty-string (str stream)
  (let ((*standard-output* stream)
	(word (make-array 10 :adjustable t 
			  :fill-pointer 0 :element-type 'character))
	spaceflag)
    (loop for c across str
	  unless (find c *space-chars*)
	  do (vector-push-extend c word) (setf spaceflag nil)
	  else do (unless spaceflag 
		    (put-word word stream)
		    (setf spaceflag t)
		    (when (char= c #\Newline) 
		      (setf spaceflag nil) (newline stream))
		    (adjust-array word 10 :fill-pointer 0)))
    (unless (zerop (length word)) (put-word word stream))))
   	  
(defun sprint (str &rest args)
  (pretty-string (apply #'format nil (ignore-newlines str) args) *outstream*)
  nil)

(defun addword (word)
  (let ((word (string-downcase word)))
    (multiple-value-bind (num ex) (gethash word *dictionary*)
      (declare (ignore num))
      (unless ex (setf (gethash word *dictionary*) (incf *dict-index*))))))

(define-condition nosuchword (error)
  ((word :initarg :word :reader nosuchword-word))
  (:report (lambda (condition stream)
	     (format stream "No such word in dictionary: ~a" 
		     (nosuchword-word condition)))))

(defun word2dic (word)
  (let ((word (string-downcase word)))
    (multiple-value-bind (num ex) (gethash word *dictionary*)
      (if ex num (error 'nosuchword :word word)))))

(defun addword2dic (word)
  (let ((word (string-downcase word)))
    (multiple-value-bind (num ex) (gethash word *dictionary*)
      (if ex num (setf (gethash word *dictionary*) (incf *dict-index*))))))

(defun split-to-words (string) 
  "Returns a list of words in string"
  (assert (stringp string))
  (loop 
   with lst = nil
   with curword = ""
   for x across string
   if (find x *space-chars*) 
   do (unless (zerop (length curword)) (push curword lst))
   (setf curword "")
   else do (setf curword (format nil "~a~a" curword x))
   finally (unless (zerop (length curword)) (push curword lst))
   (return (reverse lst))))

(defgeneric parser (obj words) 
  (:documentation 
   "Parser for object - returns a number between 0 and 1 indicating
   how close the guess is."))

(defgeneric initnames (obj names)
  (:documentation "Init list of names for object"))

(defgeneric read-property (obj property &rest args)
  (:documentation "Read property of object"))

(defgeneric initflags (obj)
  (:documentation "Adds default flags for object")
  (:method-combination append :most-specific-last))

;;(defsyn rp #'read-property)
;;(defmacro rp (&rest args) `(read-property ,@args))

(defclass abstractobject ()
  ((name :initarg :name :initform "object" :accessor name
	 :documentation "Name of the object")
   (names :initform nil :reader names
	  :documentation "List of dict-words for the parser")
   (parent :initarg :parent :initform nil :reader parent
	   :documentation "Parent of object")
   (children :initform nil :reader children
	     :documentation "Children of object")
   (flags :initarg :flags :initform nil :accessor flags
	  :documentation "Flags of object")))

(defmethod initialize-instance :after ((this abstractobject) &key)
  (setf (slot-value this 'flags) (combine-flags (initflags this)))
  (push this *allobjects*))

(defmethod initflags append ((obj abstractobject))
   (declare (ignore obj))
    (list :object))

(defun flag-compare (flag1 flag2)
  "Tests whether flag2 unsets flag1"
  (let ((fl1 (symbol-name flag1))
	(fl2 (symbol-name flag2)))
    (and (char= (aref fl2 0) #\~) (string= fl1 (subseq fl2 1)))))

(defun combine-flags (flaglist)
  (loop for fl in flaglist
	if (char= (aref (symbol-name fl) 0) #\~) 
	   do (setq set (nset-difference set (list fl) :test #'flag-compare))
	else collect fl into set
	finally (return set)))

(defun add-flags (obj &rest flags)
  (setf (flags obj) (combine-flags (append (flags obj) flags)))) 

(defun give (obj &rest flags) ;;Synonim to add-flags
  (setf (flags obj) (combine-flags (append (flags obj) flags))))

(defun has (obj &rest flags)
  (subsetp flags (flags obj)))

(defun hasnt (obj &rest flags)
  (not (intersection flags (flags obj))))
  ;(not (subsetp flags (flags obj))))

(defmethod parser ((obj abstractobject) words) 
  "Default parser"
  (when (zerop (length words)) (return-from parser 0))
  (let ((words1 (remove-duplicates words)))
    (/ (loop 
	for word in words
	counting (member word (names obj)))
       (length words1))))

(defmethod initnames ((obj abstractobject) names)
  "Initialise names for object"
  (setf (slot-value obj 'names) 
	(remove-duplicates (mapcar #'addword2dic names))))

(defun addnames (obj names)
  (initnames obj (remove-duplicates 
		  (append (names obj) (mapcar #'addword2dic names)))))

(defun eval-err (value type)
  (error "~S cannot be evaluated as ~a." value type))

(defun read-property-string (value &rest args)
  (cond ((stringp value) value)
	((not value) "")
	((functionp value) 
	 (let ((res (apply value args)))
	   (read-property-string res args))) 
	((numberp value) (format nil "~a" value))
	((and (typep value 'abstractobject)
	      (slot-exists-p value 'name)
	      (stringp (name value))) (name value))
	(t (eval-err value "string"))))

(defun read-property-number (value &rest args)
  (cond ((numberp value) value)
	((not value) 0)
	((functionp value) 
	 (let ((res (apply value args)))
	   (read-property-number res args))) 	
	(t (eval-err value "number"))))

(defun read-property-integer (value &rest args)
  (cond ((integerp value) value)
	((not value) 0)
	((functionp value) 
	 (let ((res (apply value args)))
	   (read-property-integer res args))) 	
	((stringp value) (parse-integer value :junk-allowed t))
	(t (eval-err value "integer"))))
  
(defun read-property-object (value &rest args)
  (cond ((typep value 'abstractobject) value)
	((not value) nil) 
	((functionp value)
	 (let ((res (apply value args)))
	   (read-property-object res args)))
	((symbolp value)
	 (let ((res (symbol-value value)))
	   (read-property-object res args)))
	((stringp value) (sprint "~a~%" value) (values value t))
	(t (eval-err value "object"))))

(defmacro exec (func (&rest args) &key str)
  (with-gen-syms (tmp)
    `(let ((,tmp (apply #',func (list ,@args))))
      ,(unless str 
	       `(when (stringp ,tmp) (sprint ,tmp) (newline *outstream*)))
      (values ,tmp t))))

(defun exec* (func args &key str)
  (let ((args (if (listp args) args (list args))))
  (let ((tmp (apply func args)))
    (unless str (when (stringp tmp) (sprint tmp) (newline *outstream*)))
    (values tmp t))))

(defun read-property-execute (value &rest args)
  (cond ((functionp value) (exec* value args))
	((not value) nil)
	((stringp value) (sprint "~a~%" value) (values value t))
	(t value)))

(defun read-property-list (value &rest args)
  (cond ((listp value) value)
	((functionp value)
	 (let ((res (apply value args)))
	   (if (listp res) res (list res))))
	(t (list value))))

(defun read-property-other (value &rest args)
  (declare (ignore args)) value)

(defmethod read-property ((obj abstractobject) property &rest args)
  "default read-property"
  (case property
    (name (apply #'read-property-string (slot-value obj property) args))
    (description (apply #'read-property-string (slot-value obj property) args))
    (article (apply #'read-property-string (slot-value obj property) args))
    (glance (apply #'read-property-string (slot-value obj property) args))
    (in-scope (apply #'read-property-list (slot-value obj property) args))
    (t (slot-value obj property))))


(defun type-keywordp (obj)
  (and (symbolp obj)
       (cdr (assoc (symbol-name obj) 
		   (mapcar #'(lambda (s) (cons (symbol-name s) s)) 
			   '(string number integer object function))))))      

(defun parse-prop (prop)
  (let* ((p1 (first prop))
	(p2 (second prop))
	(p3 (third prop))
	(p4 (fourth prop))
	(ggg (type-keywordp p2)))
    (case (length prop)
      (1 (list p1 nil nil))
      (2 (if ggg (list p1 ggg) (list p1 nil p2)))
      (3 (if ggg (list p1 ggg p3) (list p1 nil p2 p3)))
      (4 (list p1 p2 p3 p4)))))

(defun prop-process1 (name type &optional initform (documentation ""))
  (unless initform 
    (setf initform (case type 
		     (string "")
		     (number 0)
		     (integer 0)
		     (object nil)
		     (function nil)
		     (list nil)
		     (t nil))))
  `(,name :initarg ,(as-keyword name) :accessor ,name :initform ,initform
    :documentation ,documentation))

(defun prop-process2 (name type &rest stuff)
  (declare (ignore stuff))
  (unless type (return-from prop-process2 nil))
  `((,name) (apply 
	     ,(case type
		    (string #'read-property-string)
		    (number #'read-property-number)
		    (integer #'read-property-integer)
		    (object #'read-property-object)
		    (function #'read-property-execute)
		    (list #'read-property-list)
		    (t #'read-property-other))
	     (slot-value obj property) args)))

(defgeneric before (obj)
  (:method-combination or))

(defgeneric after (obj)
  (:method-combination or))
   
(defmethod before or (obj) (declare (ignore obj)) nil)

(defmethod after or (obj) (declare (ignore obj)) nil)

(defun generate-before-after (name beforeform afterform)
  (let ((bef (cdr beforeform)) (aft (cdr afterform)))
    `(,(when bef `(defmethod before or ((self ,name))
		   (declare (ignorable self))
		   (case *action* ,@bef)))
      ,(when aft `(defmethod after or ((self ,name))
		   (declare (ignorable self))
		   (case *action* ,@aft))))))

(defmacro ifclass (name (&rest classes) &rest options)
  "Macro for generating IF classes"
  (let (beforeform afterform)
  (multiple-value-bind (proplist flaglist)
      (loop for opt in options
	    for word = (car opt)
	    ;do (format t "~a:~a:~a " opt word (eql word 'has))
	    if (eql word 'has) collect opt into fllist
	    else if (eql word 'before) do (setf beforeform opt)
	    else if (eql word 'after) do (setf afterform opt)
	    else collect opt into prlist
	    finally (return (values prlist fllist)))
    `(progn
      ;;(declare (ignorable self))
      (defclass ,name ,(or classes '(abstractobject))
	,(loop for prop in proplist
	       when (apply #'prop-process1 (parse-prop prop)) collect it))
      (defmethod read-property ((obj ,name) property &rest args)
	(declare (ignorable args))
	(case property
	  ,@(loop for prop in proplist 
		  when (apply #'prop-process2 (parse-prop prop)) collect it)
	  (t (call-next-method))))
      (handler-bind ((warning #'ignore-warning))
      (defmethod initflags append ((obj ,name))
	(declare (ignore obj))
	(list ,@(loop for fl in flaglist appending (cdr fl))))
      ,@(generate-before-after name beforeform afterform)))))) 

(defun move (obj1 obj2)
  (let (objt)
    (when (setf objt (parent obj1))
      (setf (slot-value objt 'children) 
	    (remove obj1 (slot-value objt 'children)))
      (setf (slot-value obj1 'parent) nil))
    (when obj2
      (pushnew obj1 (slot-value obj2 'children))
      (setf (slot-value obj1 'parent) obj2))))

(defmacro rmv (obj1)
  `(move ,obj1 nil))

(defun destroy (obj)
  "Destroy the object, like, totally!"
  (setf *allobjects* (remove obj *allobjects*))
  (rmv obj))

(defmacro object (intname (&rest classes) &rest options)
  (multiple-value-bind (extname parent namelist proplist flaglist)
      (loop with extname = ""
	    with parent = nil
	    for word = nil
	    for opt in options
	    if (listp opt) do (setq word (car opt))
	    else if (stringp opt) do (setq extname opt)
	         else do (setq parent opt)
	    if word
	       if (eql word 'has) collect opt into fllist
	       else if (eql word 'name) collect opt into nmlist
	       else collect opt into prlist
	    finally (return (values extname parent nmlist prlist fllist)))
    (let (listwords) 
      (unless (or namelist (endp (setq listwords (split-to-words extname))))
	(setf namelist `((name ,@listwords)))))
    (with-gen-syms (this) ;other)
    `(progn 
      (ifclass ,intname ,classes ,@proplist ,@flaglist)
      (defmethod initialize-instance :after ((,this ,intname) &key)
	(setf (slot-value ,this 'name) ,extname)
	(initnames ,this ',(loop for nm in namelist appending (cdr nm)))
	(move ,this ,parent))
      ;;(when (boundp ',intname)
      ;;  (let ((,other (symbol-value ',intname)))
      (defparameter ,intname (make-instance ',intname))))))

(defmacro defaction (name (&rest args) &body body)
  (let ((doc (car body)) (body2 (cdr body)))
    (unless (and (stringp doc) body2)  (setf doc "") (setf body2 body))    
    `(progn 
      (defgeneric ,name ,args (:documentation ,doc))
      (defmethod ,name ,args (declare (ignorable ,@args)) ,@body2))))

;;(defun add-to-end (list value)
;;  (do ((p list (cdr p)))
;;      ((endp (cdr p)) (rplacd p (if (listp value) value (list value))))))

(defun add-to-end (plist value)
  (setf (patternlist-value plist)
	(append (patternlist-value plist) 
		(if (listp value) value (list value)))))

(defun add-to-start (plist value)
  (setf (patternlist-value plist)
	(append (if (listp value) value (list value)) 
		(patternlist-value plist))))

(defun add-verb-pattern (verb plist)
  (setf (gethash verb *verbs*) plist))

(defun extend-verb-pattern (verb pattern)
  (add-to-end (gethash verb *verbs*) pattern))

(defun extend-verb-pattern-first (verb pattern)
  (add-to-start (gethash verb *verbs*) pattern))

(defun verb (&rest args)
  (multiple-value-bind (namelist pattern)
      (loop for cons on args
	    unless (stringp (car cons)) 
	    return (values (ldiff args cons) cons))
    (let ((newpattern (make-patternlist :value pattern)))
      (mapcar #'(lambda (name) 
		  (add-verb-pattern (addword2dic name) newpattern))
	      namelist))))

(defun extend-verb (name &rest pattern)
  (extend-verb-pattern (word2dic name) pattern))

(defun extend-verb-first (name &rest pattern)
  (extend-verb-pattern-first (word2dic name) pattern))

(defun extend-verb-only (&rest args)
  (multiple-value-bind (namelist pattern)
      (loop for cons on args
	    unless (stringp (car cons)) 
	    return (values (ldiff args cons) cons))
    (let* ((name (car namelist))
	   (dicname (word2dic name))
	   (patt (make-patternlist 
		  :value (patternlist-value (gethash dicname *verbs*)))))
      (mapcar #'(lambda (name) 
		  (add-verb-pattern (addword2dic name) patt))
	      namelist)
      (extend-verb-pattern dicname pattern))))

(defun extend-verb-only-first (&rest args)
  (multiple-value-bind (namelist pattern)
      (loop for cons on args
	    unless (stringp (car cons)) 
	    return (values (ldiff args cons) cons))
    (let* ((name (car namelist))
	   (dicname (word2dic name))
	   (patt (make-patternlist 
		  :value (patternlist-value (gethash dicname *verbs*)))))
      (mapcar #'(lambda (name) 
		  (add-verb-pattern (addword2dic name) patt))
	      namelist)
      (extend-verb-pattern-first dicname pattern))))

;;The pattern has the following format:
;;(...list of tokens... -> ACTION ORDER)
;;ACTION - name of action (no action by default)
;;ORDER - function that given a list shuffles it according to required order 
;; (id by default)

  
(defun parse-pattern (pattern)
  (loop for p on pattern
	if (eql (car p) '->) 
	   return (values (ldiff pattern p) 
			  (second p) (third p)) 
	finally (return (values pattern nil nil))))   

(defmacro deftoken (name &body body)
  "The body should contain a function that, given wordlist and tokenlist
   returns state of success, the result and remaining words. Each token 
   consumes one arg but can use every remaining one for additional
   information (e.g. :multiinside can see the next tokens to determine
   whether an object is inside another object"
  `(setf (gethash ,name *tokens*)
    #'(lambda (wordlist tokenlist)
	(declare (ignorable wordlist tokenlist))
	,@body)))

(defun string== (str1 str2)
  "Case-insensitive string="
  (string= (string-downcase str1) (string-downcase str2)))

(defun shuffle (list order)
  (if order (funcall order list) list))

(defun matchp (words pattern &aux arglist) 
  "Matches string against pattern and returns action and args on success"
  (multiple-value-bind (tokenlist action order) (parse-pattern pattern)
    (loop named mainloop
          for tokens on tokenlist
	  for token = (car tokens)
	  when (stringp token)
	       do (when (loop with spl = (split-to-words token)
			      while (and words spl 
					 (string== (car words) (car spl)))
			      do (pop words) (pop spl)
			      finally (return spl))
		    (return-from mainloop nil))
	  else when (listp token) 
	       do (let* ((tokenfun (gethash (first token) *tokens*))
			 (tpar (second token))
			 (*tokenpar* (if (functionp tpar) 
					 (funcall tpar) tpar)))
		    (declare (ignorable *tokenpar*))
		    (unless tokenfun (return-from mainloop nil))
		    (multiple-value-bind (success result remwords)
			(funcall tokenfun words tokens)
		      (unless success (return-from mainloop nil))
		      (when result (push result arglist))
		      (setf words remwords)))
	  else do (let ((tokenfun (gethash token *tokens*)))
		    (unless tokenfun (return-from mainloop nil))
		    (multiple-value-bind (success result remwords)
			(funcall tokenfun words tokens)
		      (unless success (return-from mainloop nil))
		      (push result arglist)
		      (setf words remwords)))
	  finally (if (endp words) 
		      (return-from mainloop 
			(values action 
				(shuffle (reverse arglist) order))) 
		      (return-from mainloop nil)))))


(defun in (obj1 obj2)
  (eql (parent obj1) obj2))

(defmacro objectloop ((&whole alltest iter &rest test) &body body)
  "Iterates over objects satisfying test. Use !last! as an indicator
   that the loop is reaching its end"
  (with-gen-syms (iterlist lastone)
    (if test
	(let ((iterator (first test)))
	  `(let ((,iterlist
		  (case ',iter
		    (eql (list ,(second test)))
		    (in (children ,(second test)))
		    (member ,(second test))
		    (t (remove-if 
			#'(lambda (,iterator) (not ,alltest)) *allobjects*)))))
	    (let ((,lastone (car (last ,iterlist))))
	      (dolist (,iterator ,iterlist) 
		(let ((!last! (eql ,iterator ,lastone)))
		  (declare (ignorable !last!))
		  ,@body)))))
	`(dolist (,iter *allobjects*) ,@body))))

(defmacro provides (obj slot)
  `(slot-exists-p ,obj ',slot))

(defun unknown-verb (word)
  (format nil "No such verb: \"~a\"" word))

;(defun turn-passing (&optional time) 
;  (declare (ignore time))
;  "This is a stub" nil)

(declaim (ftype function turn-passing))

(defun run-action-after (obj)
  (setf *after* t)
  (when (and *args* (exec* #'after obj))
    (setf *after* nil)
    (return-from run-action-after nil))
  ;;React after?
  t)

(defun run-action (action args &key (time 1))
  (unless (listp args) (setf args (list args)))
  (setf *after* nil)
  (let ((*action* action)
	(*args* args)
	(*noun* (first args))
	(*second* (second args)))
    ;;React before?
    ;;Run before?
    (when (and args (exec* #'before (car args))) 
      (return-from run-action nil))
    ;;Normal action
    (exec* action args)
    (unless (zerop time) (turn-passing time))
    t))

(defun parse-command (string)
  "Parse command and execute appropriate action"
  (destructuring-bind (verb . words) (split-to-words string)    
    (handler-bind ((nosuchword #'(lambda (condition)
				   (exec unknown-verb 
					 ((nosuchword-word condition))) 
				   (return-from parse-command nil))))
      (let* ((dverb (word2dic verb))
	     (plist (gethash dverb *verbs*))
	     (patterns (when plist (patternlist-value plist))))
	(unless plist (signal 'nosuchword :word verb))
	;;(print words) (print patterns)
	(loop for pat in patterns
	      when (multiple-value-bind (action args) (matchp words pat)
		     (when action (run-action action args) t))
	      return nil
	      finally (sprint "Sorry, I can't parse that.~%"))))))
		        
    